@@ -9,6 +9,7 @@
 #include <libgnomeprint/gnome-font.h>
 #include <gnome-xml/tree.h>
 #include <gnome-xml/parser.h>
+#include <gnome-xml/xmlmemory.h>
 
 int _gnome_print_debug = -1;
 
@@ -52,9 +53,20 @@ static void gnome_font_class_init (GnomeFontClass *klass);
 static void gnome_font_init (GnomeFont *font);
 
 static void gnome_font_finalize (GtkObject *object);
+static void gnome_font_set_arg (GtkObject *o, GtkArg *arg, guint arg_id);
+static void gnome_font_get_arg (GtkObject *object, GtkArg *arg, guint arg_id);
 
 static GtkObjectClass *parent_class = NULL;
 
+/* The arguments we take */
+enum {
+	ARG_0,
+	ARG_ASCENDER,
+	ARG_DESCENDER,
+	ARG_UNDERLINE_POSITION,
+	ARG_UNDERLINE_THICKNESS
+};
+
 /* These are just the glyphs in the standard encoding. I need to add
    the rest of the glyphs, accessible through nonstandard
    encodings. */
@@ -468,7 +480,7 @@ xmlGetValue (xmlNodePtr node, const char *name)
 	if (ret != NULL) {
 		char *gret;
 		gret = g_strdup (ret);
-		free (ret);
+		xmlFree (ret);
 		return gret;
 	}
 	child = node->childs;
@@ -482,7 +494,7 @@ xmlGetValue (xmlNodePtr node, const char *name)
 			if (ret != NULL)
 			  {
 			    char *ret2 = g_strdup(ret);
-			    free(ret);
+			    xmlFree(ret);
 			    return (ret2);
 			  }
 		}
@@ -716,7 +728,7 @@ gnome_font_load_afm (GnomeFontMap *fontmap_entry)
   afm_f = fopen (fontmap_entry->afm_fn, "r");
   if (afm_f != NULL)
     {
-      status = parseFile (afm_f, &fi, P_M | P_P);
+      status = parseFile (afm_f, &fi, P_G | P_M | P_P);
 #ifdef VERBOSE
       debugmsg (_("status loading %s = %d\n"), fontmap_entry->afm_fn, status);
 #endif
@@ -729,6 +741,14 @@ gnome_font_load_afm (GnomeFontMap *fontmap_entry)
 	  GnomeFontKernPair *ktab;
 	  int ktabsize;
 
+	  fontmap_entry->ascender = fi->gfi->ascender;
+	  /* This is negated because adobe fonts use a negative number
+             for descenders below the baseline, but most people expect
+             a positive number for descenders below the line. */
+	  fontmap_entry->descender = -fi->gfi->descender;
+	  fontmap_entry->underline_position = fi->gfi->underlinePosition;
+	  fontmap_entry->underline_thickness = fi->gfi->underlineThickness;
+
 	  widths = g_malloc (256 * sizeof (int));
 	  fontmap_entry->widths = widths;
 
@@ -899,7 +919,19 @@ gnome_font_class_init (GnomeFontClass *class)
 
   parent_class = gtk_type_class (gtk_object_get_type ());
 
+  
+  gtk_object_add_arg_type ("GnomeFont::ascender", GTK_TYPE_DOUBLE, 
+			   GTK_ARG_READABLE, ARG_ASCENDER);
+  gtk_object_add_arg_type ("GnomeFont::descender", GTK_TYPE_DOUBLE, 
+			   GTK_ARG_READABLE, ARG_DESCENDER);
+  gtk_object_add_arg_type ("GnomeFont::underline_position", GTK_TYPE_DOUBLE, 
+			   GTK_ARG_READABLE, ARG_UNDERLINE_POSITION);
+  gtk_object_add_arg_type ("GnomeFont::underline_thickness", GTK_TYPE_DOUBLE, 
+			   GTK_ARG_READABLE, ARG_UNDERLINE_THICKNESS);
+
   object_class->finalize = gnome_font_finalize;
+  object_class->set_arg = gnome_font_set_arg;
+  object_class->get_arg = gnome_font_get_arg;
 
   class->n_fonts = 0;
   class->fontmap = NULL;
@@ -945,6 +977,135 @@ gnome_font_new (const char *name, double size)
   return font;
 }
 
+GnomeFont *
+gnome_font_new_from_full_name (const char *name)
+{
+  char *copy;
+  char *str_size;
+  double size;
+  GnomeFont *font;
+
+  g_return_val_if_fail(name != NULL, NULL);
+
+  copy = g_strdup(name);
+  str_size = strrchr(copy, ' ');
+  if ( str_size ) 
+    {
+      *str_size = 0;
+      str_size ++;
+    size = atof (str_size);
+    } 
+  else 
+    size = 10;
+
+  font = gnome_font_new( copy, size );
+  g_free(copy);
+  return font;
+}
+
+static void
+gnome_font_set_arg (GtkObject *o, GtkArg *arg, guint arg_id)
+{
+	GnomeFont *gnome_font;
+
+	gnome_font = GNOME_FONT (o);
+	
+	switch (arg_id){
+	default:
+	  break;
+	}
+}
+
+static void
+gnome_font_get_arg (GtkObject *object, GtkArg *arg, guint arg_id)
+{
+	GnomeFont *gnome_font;
+
+	gnome_font = GNOME_FONT (object);
+
+	switch (arg_id) {
+	case ARG_ASCENDER:
+	  GTK_VALUE_DOUBLE (*arg) = gnome_font_get_ascender(gnome_font);
+	  break;
+	case ARG_DESCENDER:
+	  GTK_VALUE_DOUBLE (*arg) = gnome_font_get_descender(gnome_font);
+	  break;
+	case ARG_UNDERLINE_POSITION:
+	  GTK_VALUE_DOUBLE (*arg) = gnome_font_get_underline_position(gnome_font);
+	  break;
+	case ARG_UNDERLINE_THICKNESS:
+	  GTK_VALUE_DOUBLE (*arg) = gnome_font_get_underline_thickness(gnome_font);
+	  break;
+	default:
+	  arg->type = GTK_TYPE_INVALID;
+	  break;
+	}
+}
+
+/**
+ * gnome_font_get_ascender:
+ * @font: the GnomeFont to operate on
+ *
+ * This works with the standard Adobe encoding and without kerning or
+ * ligatures. When the text libs get written, this function will be
+ * deprecated.
+ *
+ * Returns: The ascender of the font.
+ */
+double
+gnome_font_get_ascender (const GnomeFont *font)
+{
+  return font->fontmap_entry->ascender * font->scale;
+}
+
+/**
+ * gnome_font_get_descender:
+ * @font: the GnomeFont to operate on
+ *
+ * This works with the standard Adobe encoding and without kerning or
+ * ligatures. When the text libs get written, this function will be
+ * deprecated.
+ *
+ * Returns: The descender of the font.
+ */
+double
+gnome_font_get_descender (const GnomeFont *font)
+{
+  return font->fontmap_entry->descender * font->scale;
+}
+
+/**
+ * gnome_font_get_underline_position:
+ * @font: the GnomeFont to operate on
+ *
+ * This works with the standard Adobe encoding and without kerning or
+ * ligatures. When the text libs get written, this function will be
+ * deprecated.
+ *
+ * Returns: The underline position of the font.
+ */
+double
+gnome_font_get_underline_position (const GnomeFont *font)
+{
+  return font->fontmap_entry->underline_position * font->scale;
+}
+
+/**
+ * gnome_font_get_underline_thickness:
+ * @font: the GnomeFont to operate on
+ *
+ * This works with the standard Adobe encoding and without kerning or
+ * ligatures. When the text libs get written, this function will be
+ * deprecated.
+ *
+ * Returns: The underline thickness of the font.
+ */
+double
+gnome_font_get_underline_thickness (const GnomeFont *font)
+{
+  return font->fontmap_entry->underline_thickness * font->scale;
+}
+
 /* Find the closest weight matching the family name, weight, and italic
    specs. Return the unsized font. */
 GnomeFontUnsized *
@@ -1236,6 +1397,12 @@ gnome_font_unsized_get_pfa (const GnomeFontUnsized *font)
   return flat;
 }
 
+char *
+gnome_font_get_full_name (const GnomeFont *font)
+{
+  return g_strdup_printf("%s %f", gnome_font_get_name(font), font->size);
+}
+
 char *
 gnome_font_get_pfa (const GnomeFont *font)
 {
@@ -1722,20 +1889,23 @@ static GdkFont *find_best_x_font (const char *unsized_name,
   /*char *test_string = "abcdefghijklmnopqrstuvwxyz";*/
   /*char *test_string = "snider";*/
   const char *test_string = "oooowwww";
+#if 0
   int length = strlen (test_string);
 
   double perfect_width; /* the printer-font width of the sample string */
   double max_width; /* we must end up below this size */
 
+  int direction; /* either +1 or -1 -- are we looking for a bigger font
+		    or a smaller one? */
+
   /* information about the best match found so far */
-  int current_size; /* point size */
   int current_width; /* width of rendered sample string */
   int current_diff; /* difference between current_width and perfect_width */
+#endif
+  int current_size; /* point size */
   char *current_x_name; /* the x-style name of the current font */
   GdkFont *current_font; /* the current font */
 
-  int direction; /* either +1 or -1 -- are we looking for a bigger font
-		    or a smaller one? */
 #if 0
   g_return_val_if_fail (unsized_name, NULL);
   g_return_val_if_fail (gnome_font, NULL);
