@@ -5,9 +5,19 @@
 
 #ifdef __cplusplus
 extern "C" {
-#pragma }
 #endif /* __cplusplus */
 
+/* GnomeFont - A gnome font object.
+ *
+ * The following arguments are available:
+ *
+ * name		       type		read/write	description
+ * --------------------------------------------------------------------------------
+ * ascender            double           R               maximum ascender for the font
+ * descender           double           R               maximum descender for the font
+ * underline_position  double           R               underline position for the font
+ * underline_thickness double           R               underline thickness for the font
+ */
 
 #define GNOME_TYPE_FONT		 (gnome_font_get_type ())
 #define GNOME_FONT(obj)		 (GTK_CHECK_CAST ((obj), GNOME_TYPE_FONT, GnomeFont))
@@ -84,6 +94,10 @@ struct _GnomeFontMap
   gboolean italic;
   /* above is font metric info stored in the fontmap, below is font metric
      info parsed from the afm file. */
+  int   ascender;
+  int   descender;
+  int   underline_position;
+  int   underline_thickness;
   int  *widths;
   GnomeFontKernPair *kerns;
   int num_kerns;
@@ -115,6 +129,8 @@ GnomeFont *gnome_font_new_closest (const char *family_name,
 				   gboolean italic,
 				   double size);
 
+GnomeFont *gnome_font_new_from_full_name (const char *string);
+
 /* Find the closest weight matching the family name, weight, and italic
    specs. Return the unsized font. */
 GnomeFontUnsized *
@@ -142,6 +158,9 @@ gnome_font_unsized_get_pfa (const GnomeFontUnsized *font);
 char *
 gnome_font_get_pfa (const GnomeFont *font);
 
+char *
+gnome_font_get_full_name (const GnomeFont *font);
+
 /* this routine is fast; go ahead and use it */
 double
 gnome_font_get_width (const GnomeFont *font, int ch);
@@ -156,6 +175,19 @@ gnome_font_get_width_string (const GnomeFont *font, const char *s);
 double
 gnome_font_get_width_string_n (const GnomeFont *font, const char *s, int n);
 
+/* these routines return the ascender, descender, and underline
+   information as specified by postscript.  Descender is negated since
+   most people expect a positive descender for descenders below the
+   baseline */
+double
+gnome_font_get_ascender (const GnomeFont *font);
+double
+gnome_font_get_descender (const GnomeFont *font);
+double
+gnome_font_get_underline_position (const GnomeFont *font);
+double
+gnome_font_get_underline_thickness (const GnomeFont *font);
+
 /* Return the amount of kerning for the two glyphs. */
 double
 gnome_font_kern (const GnomeFont *font, int glyph1, int glyph2);
@@ -212,5 +244,4 @@ GnomeDisplayFont *gnome_get_display_font      (const char *family,
 }
 #endif /* __cplusplus */
 
-
 #endif /* __GNOME_FONT_H__ */
