@@ -1,8 +1,9 @@
 /*
  * Metafile printer driver.
  *
- * Author:
+ * Authors:
  *    Miguel de Icaza (miguel@gnu.org
+ *    michael Zucchi <notzed@helixcode.com>
  *
  * TODO:
  *    If the GnomeFont is not found during the replay of a metafile
@@ -13,6 +14,11 @@
  *    all the fonts when the render is done.
  *
  *    Handle GNOME_META_TEXTLINE
+ *
+ *    Because each page is played back in isolation, if there are any
+ *    options that persist across showpage's, then they need to be
+ *    tracked, and stored in the page header for each page.  But I
+ *    dont think there are ...
  */
 #include <config.h>
 #include <gtk/gtk.h>
@@ -23,14 +29,34 @@
 
 static GnomePrintContextClass *parent_class = NULL;
 
-#define GNOME_METAFILE_SIGNATURE "GNOME_METAFILE-1.0"
+/* Note: this must continue to be the same length  as "GNOME_METAFILE-0.0" */
+#define GNOME_METAFILE_SIGNATURE "GNOME_METAFILE-1.1"
+#define GNOME_METAFILE_SIGNATURE_SIZE 18
 #define BLOCKSIZE 4096
 
 typedef struct {
-	char signature [sizeof (GNOME_METAFILE_SIGNATURE)];
-	int  size;
+	char signature [GNOME_METAFILE_SIGNATURE_SIZE];
+	gint32  size;
 } GnomeMetaFileHeader;
 
+/* this is used to quickly scan pages through a metafile, it is
+   stored at the start of the file, and after every showpage */
+#define PAGE_SIGNATURE "PAGE"
+typedef struct {
+	char signature[4];	/* magic token for a page header */
+	gint32 size;
+} GnomeMetaPageHeader;
+
+/* ENCODED sizes maybe != structure sizes (alignment) */
+#define FILEHEADER_SIZE (GNOME_METAFILE_SIGNATURE_SIZE+4)
+#define PAGEHEADER_SIZE (4+4)
+
+struct _GnomePrintMetaPrivate {
+	int last_page;		/* offset into the buffer for the last page header */
+};
+
+static const char *decode_header (const char *data, GnomeMetaFileHeader *mh);
+
 typedef enum {
 	GNOME_META_NEWPATH,
 	GNOME_META_MOVETO,
@@ -69,14 +95,15 @@ gnome_print_meta_finalize (GtkObject *object)
 
 	meta = GNOME_PRINT_META (object);
 
-	if (meta->buffer)
-		g_free (meta->buffer);
-				      
+	g_free (meta->buffer);
+	g_free (meta->priv);
+
 	(*GTK_OBJECT_CLASS (parent_class)->finalize) (object);
 }
 
-static void
-encode_block (GnomePrintMeta *meta, int size, void const *data)
+/* returns true if there is room (grows it if there is not) */
+static gboolean
+check_room (GnomePrintMeta *meta, int size)
 {
 	int bytes_left;
 
@@ -88,24 +115,68 @@ encode_block (GnomePrintMeta *meta, int size, void const *data)
 		
 		newp = g_realloc (meta->buffer, meta->buffer_size + min_size);
 		if (!newp)
-			return;
+			return FALSE;
 
 		meta->buffer = newp;
 		meta->buffer_size = meta->buffer_size + min_size;
 	}
+	return TRUE;
+}
+
+static void
+encode_block (GnomePrintMeta *meta, int size, void const *data)
+{
+	if (!check_room(meta, size))
+		return;
 
 	memcpy (meta->buffer + meta->current, data, size);
 	meta->current += size;
 }
 
 static void
-encode_int (GnomePrintContext *pc, gint32 value)
+encode_int_header (GnomePrintMeta *pc, gint32 value)
 {
 	gint32 new = g_htonl (value);
 	
 	encode_block (GNOME_PRINT_META (pc), sizeof (gint32), &new);
 }
 
+static void
+encode_int (GnomePrintContext *pc, gint32 value)
+{
+	char *out;
+	GnomePrintMeta *meta = GNOME_PRINT_META(pc);
+
+	if (!check_room(meta, 6))
+		return;
+
+	/* we encode upto 7 bits of info in each byte, using the
+	   top bit to indicate the last byte, this makes decoding easier */
+	out = meta->buffer + meta->current;
+	if (value< (1<<7)) {
+		*out++ = value | 0x80;
+	} else if (value < (1<<14)) {
+		*out++ = value >> 7;
+		*out++ = (value & 0x7f) | 0x80;
+	} else if (value < (1<<21)) {
+		*out++ = value >> 14;
+		*out++ = (value >> 7) & 0x7f;
+		*out++ = (value & 0x7f) | 0x80;
+	} else if (value < (1<<28)) {
+		*out++ = value >> 21;
+		*out++ = (value >> 14) & 0x7f;
+		*out++ = (value >> 7) & 0x7f;
+		*out++ = (value & 0x7f) | 0x80;
+	} else {
+		*out++ = (value >> 28) & 0x7f; /* c doesn't say what happens to the sign bit, even for unsigned */
+		*out++ = (value >> 21) & 0x7f;
+		*out++ = (value >> 14) & 0x7f;
+		*out++ = (value >> 7) & 0x7f;
+		*out++ = (value & 0x7f) | 0x80;
+	}
+	meta->current = out - meta->buffer;
+}
+
 static void
 encode_double (GnomePrintContext *pc, double d)
 {
@@ -129,6 +200,35 @@ encode_six_doubles (GnomePrintContext *pc, double x1, double y1, double x2, doub
 	encode_double (pc, y3);
 }
 
+static void
+encode_header(GnomePrintMeta *pc, GnomeMetaFileHeader *mh)
+{
+	encode_block(pc, sizeof(mh->signature), mh->signature);
+	encode_int_header(pc, mh->size);
+}
+
+/* sets the size of the last page */
+static void
+close_page_header(GnomePrintMeta *pc)
+{
+	gint32 *sizeptr;
+
+	if (pc->priv->last_page != -1) {
+		/* we cannot map the page header back into memory, so poke it back in */
+		sizeptr = (gint32 *)((pc->buffer + pc->priv->last_page) + 4);
+		*sizeptr = g_htonl(pc->current - pc->priv->last_page - PAGEHEADER_SIZE);
+	}
+}
+
+static void
+encode_page_header(GnomePrintMeta *pc, GnomeMetaPageHeader *mh)
+{
+	close_page_header(pc);
+	pc->priv->last_page = pc->current;
+	encode_block(pc, sizeof(mh->signature), mh->signature);
+	encode_int_header(pc, mh->size);
+}
+
 static int
 meta_newpath (GnomePrintContext *pc)
 {
@@ -365,8 +465,18 @@ meta_textline (GnomePrintContext *pc, GnomeTextLine *line)
 static int
 meta_showpage (GnomePrintContext *pc)
 {
+	GnomeMetaPageHeader page_header = {
+		PAGE_SIGNATURE,
+		-1
+	};
+	GnomePrintMeta *meta = GNOME_PRINT_META(pc);
+
 	encode_int (pc, GNOME_META_SHOWPAGE);
-	GNOME_PRINT_META (pc)->pages++;
+
+	/* assume another page is coming */
+	encode_page_header(meta, &page_header);
+
+	meta->pages++;
 	return 0;
 }
 
@@ -377,7 +487,8 @@ meta_close (GnomePrintContext *pc)
 
 	encode_int (pc, GNOME_META_CLOSE);
 
-	((GnomeMetaFileHeader *)meta->buffer)->size = meta->buffer_size;
+	/* we cannot just map the header to memory, so poke it manually */
+	*((gint32 *)(meta->buffer+GNOME_METAFILE_SIGNATURE_SIZE)) = g_htonl(meta->buffer_size);
 
 	return 0;
 }
@@ -430,12 +541,21 @@ gnome_print_meta_class_init (GnomePrintMetaClass *class)
 static void
 gnome_print_meta_init (GnomePrintMeta *meta)
 {
+	GnomeMetaFileHeader header = {
+		GNOME_METAFILE_SIGNATURE,
+		-1
+	};
+	GnomeMetaPageHeader page_header = {
+		PAGE_SIGNATURE,
+		-1
+	};
+
 	meta->buffer_size = BLOCKSIZE;
 	meta->buffer = g_malloc (meta->buffer_size);
-
-	strcpy (meta->buffer, GNOME_METAFILE_SIGNATURE);
-	meta->current = sizeof (GnomeMetaFileHeader);
-	((GnomeMetaFileHeader *)meta->buffer)->size = -1;
+	meta->priv = g_malloc0 (sizeof(GnomePrintMetaPrivate));
+	meta->priv->last_page = -1;
+	encode_header(meta, &header);
+	encode_page_header(meta, &page_header);
 }
 
 /**
@@ -495,15 +615,16 @@ GnomePrintMeta *
 gnome_print_meta_new_from (const void *data)
 {
 	GnomePrintMeta *meta;
-	const GnomeMetaFileHeader *mf = data;
+	GnomeMetaFileHeader header;
 	int len;
 
 	g_return_val_if_fail (data != NULL, NULL);
 
-	if (strcmp ((const char *)data, GNOME_METAFILE_SIGNATURE) != 0)
+	decode_header(data, &header);
+	if (strncmp (header.signature, GNOME_METAFILE_SIGNATURE, GNOME_METAFILE_SIGNATURE_SIZE) != 0)
 		return NULL;
 
-	len = mf->size;
+	len = header.size;
 	meta = gtk_type_new (gnome_print_meta_get_type ());
 
 	if (meta->buffer_size < len){
@@ -533,12 +654,14 @@ gnome_print_meta_new_from (const void *data)
 int
 gnome_print_meta_access_buffer (GnomePrintMeta *meta, void **buffer, int *buflen)
 {
+	char *p;
+
 	g_return_val_if_fail (meta != NULL, 0);
 	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), 0);
 
-	((GnomeMetaFileHeader *)meta->buffer)->size = meta->buffer_size;
-	*buffer = meta->buffer;
-	*buflen = meta->buffer_size;
+	p = *buffer = meta->buffer;
+	*((gint *)(p + FILEHEADER_SIZE)) = g_htonl(meta->current);
+	*buflen = meta->current;
 
 	return 1;
 }
@@ -557,22 +680,26 @@ gnome_print_meta_access_buffer (GnomePrintMeta *meta, void **buffer, int *buflen
 int
 gnome_print_meta_get_copy (GnomePrintMeta *meta, void **buffer, int *buflen)
 {
+	char *p;
+
 	g_return_val_if_fail (meta != NULL, 0);
 	g_return_val_if_fail (GNOME_IS_PRINT_META (meta), 0);
 
-	*buffer = g_malloc (meta->buffer_size);
+	p = *buffer = g_malloc (meta->current);
 	if (*buffer == NULL)
 		return 0;
 
-	memcpy (*buffer, meta->buffer, meta->buffer_size);
-	((GnomeMetaFileHeader *)*buffer)->size = meta->buffer_size;
-	*buflen = meta->buffer_size;
+	memcpy (p, meta->buffer, meta->buffer_size);
+	*((gint *)(p + FILEHEADER_SIZE)) = g_htonl(meta->current);
+	*buflen = meta->current;
 
 	return 1;
 }
 
+/* less compact, but gauranteed size (just makes things easier to encode
+   fixed size data like headers) */
 static const char *
-decode_int (const char *data, gint32 *dest)
+decode_int_header (const char *data, gint32 *dest)
 {
 	gint32 nint;
 	memcpy (&nint, data, sizeof (gint32));
@@ -580,6 +707,21 @@ decode_int (const char *data, gint32 *dest)
 	return data + sizeof (gint32);
 }
 
+static const char *
+decode_int (const char *data, gint32 *dest)
+{
+	gint32 value=0, v;
+
+	/* until we get the last byte, keep decoding 7 bits at a time */
+	while ( ((v = *data++) & 0x80) == 0) {
+		value |= v;
+		value <<= 7;
+	}
+	value |= (v&0x7f);
+	*dest = value;
+	return data;
+}
+
 static const char *
 decode_double (const char *data, double *dest)
 {
@@ -592,7 +734,7 @@ decode_double (const char *data, double *dest)
 }
 
 static const char *
-decode_image (const char *data, GnomePrintContext *dest, int bytes_per_pixel, gboolean do_output)
+decode_image (const char *data, GnomePrintContext *dest, int bytes_per_pixel)
 {
 	void *buffer;
 	gint32 height, width;
@@ -602,61 +744,99 @@ decode_image (const char *data, GnomePrintContext *dest, int bytes_per_pixel, gb
 	data = decode_int (data, &width);
 
 	size = height * width * bytes_per_pixel;
-	if (do_output){
-		buffer = g_malloc (size);
-		memcpy (buffer, data, size);
-		
-		if (bytes_per_pixel == 1)
-			gnome_print_grayimage (dest, buffer, width, height, width);
-		else
-			gnome_print_rgbimage (dest, buffer, width, height, width * bytes_per_pixel);
-		
-		g_free (buffer);
-	}
+	buffer = g_malloc (size);
+	memcpy (buffer, data, size);
+	
+	if (bytes_per_pixel == 1)
+		gnome_print_grayimage (dest, buffer, width, height, width);
+	else
+		gnome_print_rgbimage (dest, buffer, width, height, width * bytes_per_pixel);
+	
+	g_free (buffer);
 	return data + size;
 }
 
+static const char *
+decode_block (const char *data, void *out, int len)
+{
+	memcpy(out, data, len);
+	return data+len;
+}
+
+static const char *
+decode_header (const char *data, GnomeMetaFileHeader *mh)
+{
+	data = decode_block(data, mh->signature, sizeof(mh->signature));
+	data = decode_int_header(data, &mh->size);
+	return data;
+}
+
+static const char *
+decode_page_header (const char *data, GnomeMetaPageHeader *mh)
+{
+	data = decode_block(data, mh->signature, sizeof(mh->signature));
+	data = decode_int_header(data, &mh->size);
+	return data;
+}
+
+static const char *
+locate_page_header(const char *meta_stream, int page)
+{
+	const char *data = meta_stream + FILEHEADER_SIZE;
+	GnomeMetaPageHeader page_header;
+
+	while (page>0) {
+		page--;
+		data = decode_page_header(data, &page_header);
+		/* unclosed or last page? */
+		if (page_header.size == -1)
+			return NULL;
+		data += page_header.size;
+	}
+	if (page_header.size == -1)
+		return NULL;
+	return data;
+}
+
+/*
+  from the current point in the data stream, print pages pages, -1
+  means print all
+*/
 static gboolean
-do_render (GnomePrintContext *dest, const char *data, int size, int page)
+do_render (GnomePrintContext *dest, const char *data, int size, int pages)
 {
 	const char *end = data + size;
 	double matrix [6];
 	double x1, y1, x2, y2, x3, y3;
 	double x, y;
 	double r, g, b;
-	int    current_page = 0;
-	gboolean do_output;
+	GnomeMetaPageHeader page_header;
 	
 	/*
 	 * Output if all pages were requested, or the first page was
+	 *
+	 * Note that this assumes each page is completely described
+	 * by its own data block.
 	 */
-	if (page == -1 || page == 0)
-		do_output = TRUE;
-	else
-		do_output = FALSE;
-	
 	while (data < end){
 		gint32 opcode, i;
 		
 		data = decode_int (data, &opcode);
 		switch ((GnomeMetaType) opcode){
 		case GNOME_META_NEWPATH:
-			if (do_output)
-				gnome_print_newpath (dest);
+			gnome_print_newpath (dest);
 			break;
 			
 		case GNOME_META_MOVETO:
 			data = decode_double (data, &x);
 			data = decode_double (data, &y);
-			if (do_output)
-				gnome_print_moveto (dest, x, y);
+			gnome_print_moveto (dest, x, y);
 			break;
 
 		case GNOME_META_LINETO:
 			data = decode_double (data, &x);
 			data = decode_double (data, &y);
-			if (do_output)
-				gnome_print_lineto (dest, x, y);
+			gnome_print_lineto (dest, x, y);
 			break;
 			
 		case GNOME_META_CURVETO:
@@ -667,13 +847,11 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 			data = decode_double (data, &x3);
 			data = decode_double (data, &y3);
 
-			if (do_output)
-				gnome_print_curveto (dest, x1, y1, x2, y2, x3, y3);
+			gnome_print_curveto (dest, x1, y1, x2, y2, x3, y3);
 			break;
 			
 		case GNOME_META_CLOSEPATH:
-			if (do_output)
-				gnome_print_closepath (dest);
+			gnome_print_closepath (dest);
 			break;
 
 		case GNOME_META_SETRGBCOLOR:
@@ -684,13 +862,11 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 			break;
 			
 		case GNOME_META_FILL:
-			if (do_output)
-				gnome_print_fill (dest);
+			gnome_print_fill (dest);
 			break;
 			
 		case GNOME_META_EOFILL:
-			if (do_output)
-				gnome_print_eofill (dest);
+			gnome_print_eofill (dest);
 			break;
 			
 		case GNOME_META_SETLINEWIDTH:
@@ -728,13 +904,11 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 		}
 		
 		case GNOME_META_STROKEPATH:
-			if (do_output)
-				gnome_print_strokepath (dest);
+			gnome_print_strokepath (dest);
 			break;
 			
 		case GNOME_META_STROKE:
-			if (do_output)
-				gnome_print_stroke (dest);
+			gnome_print_stroke (dest);
 			break;
 			
 		case GNOME_META_SETFONT: {
@@ -743,14 +917,14 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 			
 			data = decode_double (data, &x);
 			data = decode_int (data, &i);
+
 			name = g_malloc (i + 1);
 			memcpy (name, data, i);
 			name [i] = 0;
-			data += i;
-
 			font = gnome_font_new (name, x);
 			g_free (name);
-
+			data += i;
+			
 			if (font){
 				gnome_print_setfont (dest, font);
 				gtk_object_unref (GTK_OBJECT (font));
@@ -765,10 +939,9 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 			buffer = g_malloc (i + 1);
 			memcpy (buffer, data, i);
 			buffer [i] = 0;
-			data += i;
-			if (do_output)
-				gnome_print_show (dest, buffer);
+			gnome_print_show (dest, buffer);
 			g_free (buffer);
+			data += i;
 			break;
 		}
 		
@@ -803,21 +976,24 @@ do_render (GnomePrintContext *dest, const char *data, int size, int page)
 			break;
 			
 		case GNOME_META_GRAYIMAGE:
-			data = decode_image (data, dest, 1, do_output);
+			data = decode_image (data, dest, 1);
 			break;
 			
 		case GNOME_META_RGBIMAGE:
-			data = decode_image (data, dest, 3, do_output);
+			data = decode_image (data, dest, 3);
 			break;
 			
 		case GNOME_META_SHOWPAGE:
 			gnome_print_showpage (dest);
-			if (page == current_page){
-				return TRUE;
+
+			/* this is unused here ... */
+			data = decode_page_header(data, &page_header);
+
+			if (pages!=-1) {
+				pages--;
+				if (pages<=0)
+					return TRUE;
 			}
-			current_page++;
-			if (current_page == page)
-				do_output = TRUE;
 			break;
 
 		case GNOME_META_TEXTLINE:
@@ -844,23 +1020,23 @@ gboolean
 gnome_print_meta_render (GnomePrintContext *destination, const void *meta_stream)
 {
 	const char *data;
-	const GnomeMetaFileHeader *header;
+	GnomeMetaFileHeader header;
 
 	g_return_val_if_fail (destination != NULL, FALSE);
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (destination), FALSE);
 	g_return_val_if_fail (meta_stream != NULL, FALSE);
 
 	data = meta_stream;
-	if (strcmp (data, GNOME_METAFILE_SIGNATURE) != 0)
+	data = decode_header(data, &header);
+	if (strncmp (header.signature, GNOME_METAFILE_SIGNATURE, GNOME_METAFILE_SIGNATURE_SIZE) != 0)
 		return FALSE;
 
-	header = meta_stream;
-	if (header->size == -1){
+	if (header.size == -1){
 		g_warning ("This stream was not closed");
 		return FALSE;
 	}
 
-	return do_render (destination, data + sizeof (GnomeMetaFileHeader), header->size, -1);
+	return do_render (destination, data + FILEHEADER_SIZE + PAGEHEADER_SIZE, header.size, -1);
 }
 
 /**
@@ -899,23 +1075,29 @@ gboolean
 gnome_print_meta_render_page (GnomePrintContext *destination, const void *meta_stream, int page)
 {
 	const char *data;
-	const GnomeMetaFileHeader *header;
-	
+	GnomeMetaFileHeader header;
+	const char *page_data;
+
 	g_return_val_if_fail (destination != NULL, FALSE);
 	g_return_val_if_fail (GNOME_IS_PRINT_CONTEXT (destination), FALSE);
 	g_return_val_if_fail (meta_stream != NULL, FALSE);
 
 	data = meta_stream;
-	if (strcmp (data, GNOME_METAFILE_SIGNATURE) != 0)
+	data = decode_header(data, &header);
+	if (strncmp (header.signature, GNOME_METAFILE_SIGNATURE, GNOME_METAFILE_SIGNATURE_SIZE) != 0)
 		return FALSE;
 
-	header = meta_stream;
-	if (header->size == -1){
+	if (header.size == -1){
 		g_warning ("This printing context has not been closed");
 		return FALSE;
 	}
 
-	return do_render (destination, data + sizeof (GnomeMetaFileHeader), header->size, page);
+	page_data = locate_page_header(meta_stream, page);
+	if (page_data == NULL) {
+		g_warning("Trying to print a non-existant page");
+		return FALSE;
+	}
+	return do_render (destination, page_data+PAGEHEADER_SIZE, header.size, 1);
 }
 
 /**
