@@ -89,6 +89,9 @@ gc_set_dash (GraphicContext *ctx, int n_values, const double *values, double off
 static void
 gc_clear_bpath (GraphicContext *ctx)
 {
+	if (!ctx->bpath)
+		return;
+	
 	gnome_canvas_bpath_def_unref (ctx->bpath);
 	ctx->bpath = 0;
 }
@@ -515,6 +518,9 @@ gpp_gsave (GnomePrintContext *pc)
 
 		newctx->bpath = newpath;
 	}
+
+	if (newctx->font)
+		gtk_object_ref (GTK_OBJECT (newctx->font));
 	
 	priv->context_list = g_slist_prepend (priv->context_list, newctx);
 
@@ -722,7 +728,7 @@ gpp_show (GnomePrintContext *pc, char const *text)
 	GnomePrintPreview *pp = GNOME_PRINT_PREVIEW (pc);
 	GnomePrintPreviewPrivate *priv = pp->priv;
 	GnomeCanvasItem *item;
-	double x, y;
+	ArtPoint p;
 	Gt1LoadedFont *gt1_font;
 	double transform [6], inverse[6];
 	GraphicContext *g = gc (priv);
@@ -752,36 +758,8 @@ gpp_show (GnomePrintContext *pc, char const *text)
 	 * with the translation already done for the new
 	 * Postscript->Canvas translation
 	 */
-	x = g->bpath->bpath [0].x3;
-	y = g->bpath->bpath [0].y3;
-
-	/*
-	 * Flip the angle
-	 */
-	memcpy (inverse, g->affine, sizeof (g->affine));
-       
-	inverse [1] = -inverse [1];
-        inverse [2] = -inverse [2];
-       
-	/*
-	inverse [0] *= pp->canvas->pixels_per_unit;
-        inverse [3] *= pp->canvas->pixels_per_unit;
-	*/
-	art_affine_invert (transform, inverse);
-	x =  g->bpath->bpath [0].x3 * transform[0] + g->bpath->bpath [0].y3  * transform[2] + transform[4];
-	y =  g->bpath->bpath [0].x3 * transform[1] + g->bpath->bpath [0].y3  * transform[3] + transform[5]; 
-
-	lazy_showpage_check (priv);
-	item = gnome_canvas_item_new (
-		GNOME_CANVAS_GROUP (priv->page),
-		gnome_canvas_hacktext_get_type (),
-		"x",           x,
-		"y",           y,
-		"text",        text,
-		"font",        gt1_font,
-		"size",        g->font->size,
-		"fill_color_rgba", color_rgba (&g->color),
-		NULL);
+	p.x = g->bpath->bpath [0].x3;
+	p.y = g->bpath->bpath [0].y3;
 
 	/*
 	 * Flip the angle
@@ -789,8 +767,23 @@ gpp_show (GnomePrintContext *pc, char const *text)
 	memcpy (transform, g->affine, sizeof (g->affine));
        
 	transform [1] = -transform [1];
-	transform [2] = -transform [2];
-      
+        transform [2] = -transform [2];
+       
+	art_affine_invert (inverse, transform);
+	art_affine_point (&p, &p, inverse);
+
+	lazy_showpage_check (priv);
+	item = gnome_canvas_item_new (
+		GNOME_CANVAS_GROUP (priv->page),
+		gnome_canvas_hacktext_get_type (),
+		"x",           p.x,
+		"y",           p.y,
+		"text",        text,
+		"font",        gt1_font,
+		"size",        g->font->size,
+		"fill_color_rgba", color_rgba (&g->color),
+		NULL);
+	
 	gnome_canvas_item_affine_absolute (item, transform);
 	priv->current_page++;
 	
@@ -798,6 +791,16 @@ gpp_show (GnomePrintContext *pc, char const *text)
 	return 0;
 }
 
+static int
+gpp_setopacity (GnomePrintContext *pc, double opacity)
+{
+	GnomePrintPreview *pp = GNOME_PRINT_PREVIEW (pc);
+	GnomePrintPreviewPrivate *priv = pp->priv;
+
+	gc(priv)->color.opacity = opacity * 255;
+	return 0;
+}
+
 static int
 gpp_close (GnomePrintContext *pc)
 {
@@ -882,6 +885,7 @@ gpp_class_init (GnomePrintPreviewClass *class)
 	pc_class->rgbimage = gpp_rgbimage;
 	pc_class->textline = gpp_textline;
 	pc_class->showpage = gpp_showpage;
+	pc_class->setopacity = gpp_setopacity;
 	
 	pc_class->close = gpp_close;
 }
