@@ -101,7 +101,7 @@ profile_activate (GtkObject *item, GnomePrinterProfile *pp)
 }
 
 static void
-label_at (GtkTable *t, char *string, int col, int row)
+label_at (GtkTable *t, const char *string, int col, int row)
 {
 	GtkWidget *l = gtk_label_new (string);
 
